//----------------------------------------------------------------------------------
// File:   Sparkles.fx
// Author: Tristan Lorach
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

#include "randomtable.fxh"

#define MINSZ minSZ
#define MAXSZ maxSZ
#define FAR_CLIP near_far_scrratio.y
#define NEAR_CLIP near_far_scrratio.x
#define RATIOXY near_far_scrratio.z
#define EPSILONZ 0.1
#define SLICES 30

cbuffer changingeveryframe
{
    float4x4 proj;
    float4x4 viewProj;
    float4x4 worldViewProj;
    float4x4 worldView;
    float4x4 worldViewIT;
    float4x4 worldIT;
    float4x4 world;
    float4x4 viewIT;
    float4x4 viewI;
    float4x4 view;
};
// note : semantics used here to avoid to display them in the UI...
cbuffer almostneverchange
{
    float4 near_far_scrratio : NEAR_FAR_SCRRATIO;
    float2 DepthRTSz : DEPTHRTSZ;
};
cbuffer tweakables
{
    int     g_maxSprites = 5;
    float   minSZ < float uimin = 0.0; float uimax = 1.0; >         = 0.0;
    float   maxSZ < float uimin = 0.0; float uimax = 4.0; >         = 0.3;
    float   LODScale < float uimin = 0.0; float uimax = 35.0; >    = 30.0;
    float   Shininess < float uimin = 0.0; float uimax = 128.0; >   = 80.0;
    bool    bNoPersp = true;

    // LIGHT... we put semantic to stop the UI to display them...
    float3 lightPos : LIGHTPOS              = {20.0, 4, -10.0};
    float3 lightColor : LIGHTCOLOR          = {1.0, 1.0, 1.0};
    float3 lightColorSpec : LIGHTCOLORSPEC  = {1.0, 1.0, 1.0};
    float3 ambiLightColor : AMBILIGHTCOLOR  = {0.07, 0.07, 0.07};
    float3 surfColor : SURFCOLOR            = {0.2,0.0,0.5};
};
//
// These params are simply precomputed {cos(x),sin(x)}
// The purpose is to simulate imperfection of a micro structure for sparkling lighting of it
//
cbuffer microsurface
{
    float2 CSTable[SLICES] : CSTABLE = {{1.000000,0.000000},{1.000000,0.000000},{0.642788,0.766044},
    {0.500000,0.866025},{0.642788,0.766044},{0.766044,0.642788},{0.374607,0.927184},{0.766044,0.642788},
    {1.000000,0.000000},{0.999848,0.017452},{0.996195,0.087156},{0.567844,0.823136},{0.996195,0.087156},
    {0.999848,0.017452},{0.939693,0.342020},{0.766044,0.642788},{0.866025,0.500000},{0.999391,0.034899},
    {1.000000,0.000000},{1.000000,0.000000},{0.939693,0.342020},{1.000000,0.000000},{0.984808,0.173648},
    {0.939693,0.342020},{0.984808,0.173648},{0.996195,0.087156},{0.939693,0.342020},{0.996195,0.087156},
    {1.000000,0.000000},{1.000000,0.000000}};
};

Texture2D texture_star < String file="star1.dds"; >;
SamplerState sampler_star
{
    AddressU = Clamp;
    AddressV = Clamp;
    Filter = Min_Mag_Mip_Linear;
};

Texture2D texture_depth;
SamplerState sampler_depth
{
    AddressU = Clamp;
    AddressV = Clamp;
    Filter = Min_Mag_Mip_Point;
};

//////////////////////////////////////////////////////////////////////////////////
// STATES STATES STATES STATES STATES STATES STATES STATES STATES STATES STATES
//////////////////////////////////////////////////////////////////////////////////
DepthStencilState depthEnabled
{
    DepthEnable = true;
    DepthWriteMask = All;
    DepthFunc = Less;
};
DepthStencilState depthDisabled
{
    DepthEnable = false;
    DepthWriteMask = Zero;
    DepthFunc = Always;
};
BlendState blendOFF
{
    BlendEnable[0] = false;
};
BlendState blendON
{
    BlendEnable[0] = true;
    SrcBlend = Src_Color;
    DestBlend = One;
};
RasterizerState RStateMSAA
{
  MultisampleEnable = TRUE;
};
RasterizerState cullDisabled
{
  CullMode = None;
  MultisampleEnable = TRUE;
};
////////////////////////////////////////////////////////////////////////////////////
// STRUCTS STRUCTS STRUCTS STRUCTS STRUCTS STRUCTS STRUCTS STRUCTS STRUCTS STRUCTS
////////////////////////////////////////////////////////////////////////////////////
struct Sparkles_VSIn
{
    float3 position : position;
    float3 normal : normal;
    float2 texcoord : texcoord;
};

struct Sparkles_VSOut
{
    float4 position : SV_Position;
    float3 normal : texcoord0;
    int instanceID : texcoord1;
};

struct Sparkles_GSOut
{
    float4 position : SV_Position;
    float  alpha : color;
    float2 tc : texcoord0;
};

struct Sparkles_PSOut
{
    float4 color : SV_Target;
};

// first pass of 'car paint' is put outside of this effect because not the point of this demo
#include "CarPaint.fxh"
//////////////////////////////////////////////////////////////////////////////////
// MISC FUNCTIONS USED IN GS
//////////////////////////////////////////////////////////////////////////////////
Sparkles_VSOut barycentricWeights(const Sparkles_VSOut a, const Sparkles_VSOut b, const Sparkles_VSOut c, float3 nw)
{
    Sparkles_VSOut V;
    V.position = (nw.x * a.position + nw.y * b.position + nw.z * c.position );
    V.normal = normalize((nw.x * a.normal + nw.y * b.normal + nw.z * c.normal));
    return V;
}

float4 computeSpriteCorner(float2 dir2d, float4 position, float scale, float4 offset)
{
    float2 s;
    float4 P;
    // we want the sparkles be 2d scaled. Size not changed by perspective.
    // But you can try with perspective bNoPersp=false (and increasing maxSZ parameter to ~2.5)
    if(bNoPersp)
    {
        position /= position.w;
    }
    s = float2(1,RATIOXY)*((scale * offset.y) + offset.x);
    P.xy = dir2d*s + position.xy;
    s = float2(1,RATIOXY)*((scale * offset.w) + offset.z);
    P.x = ( dir2d.y * s.x) + P.x;
    P.y = (-dir2d.x * s.y) + P.y;
    P.zw = position.zw;
    return P;
}
float OcclusionTest(float2 P2D, float4 P)
{
    int3 iPos = int3(DepthRTSz*(P2D*float2(0.5,-0.5) + 0.5), 0);
    float d;
    d = texture_depth.Load(iPos);
    return (P.z-EPSILONZ) <= d ? 1.0 : 0.0;
}
void drawSprite(inout TriangleStream<Sparkles_GSOut> Stream, Sparkles_VSOut V, float3 Lwv, float alpha)
{
    Sparkles_GSOut output;
    if(alpha == 0.0)
      return;
    float4 Po = V.position;
    float4 Pw = mul(Po, world);
    float3 Nw = normalize(mul(float4(V.normal,0), worldIT).xyz);

    float3 lightDir = normalize(Lwv - Pw);
    float3 eyeDir = normalize(viewI[3] - Pw);
    //
    // get the second point
    //
    float4 P2wv;
    P2wv.xyz = cross(lightDir, Nw) + Pw.xyz;
    P2wv.w = 0;
    //
    // specular vector
    //
    float3 HalfV = normalize(lightDir + eyeDir);
    //
    // compute the crystal reflection
    //
    float2 NLCosSin;
    int index;
    float NDotL;
    float NDotH;

    // Note : we don't use diffuse part...
    /*NLCosSin.x =    dot(Nw, lightDir);
    NLCosSin.y =    sqrt(1-(NLCosSin.x * NLCosSin.x));
    index =         max(NLCosSin.x * SLICES, 4);
    NLCosSin =      NLCosSin * CSTable[index];
    NDotL =         NLCosSin.x + NLCosSin.y;*/

    NLCosSin.x =    max(dot(Nw, HalfV), 0);
    NLCosSin.y =    sqrt(1-(NLCosSin.x * NLCosSin.x));
    index =         max(NLCosSin.x * SLICES, 4);
    NLCosSin =      NLCosSin * CSTable[index];
    NDotH =         NLCosSin.x + NLCosSin.y;
    float scale =   pow(max(NDotH, 0), Shininess);
    //float scale =   lit(0.0/*NDotL*/, NDotH, Shininess).z; // y = diffuse, z = specular
    //
    // Projections
    //
    float4 projpos1 = mul(Pw, viewProj);
    float4 projpos2 = mul(P2wv, viewProj);
    float2 pos2d1 = projpos1.xy / projpos1.ww;
    float2 pos2d2 = projpos2.xy / projpos2.ww;
    
    scale *= OcclusionTest(pos2d1, projpos1);
    
    if(scale > 0.0)
    {
        float2 dir2d  = normalize(pos2d2 - pos2d1);

        output.alpha = alpha;
        output.position = computeSpriteCorner(dir2d, projpos1, scale, float4(-MINSZ, -MAXSZ, MINSZ, MAXSZ));
        output.tc = float2(0,1);
        Stream.Append(output);
        float4 P11 = computeSpriteCorner(dir2d, projpos1, scale, float4(MINSZ, MAXSZ, MINSZ, MAXSZ)); //P - float4(Sz.x, Sz.y, 0, 0);
        output.position = P11;
        output.tc = float2(1,1);
        Stream.Append(output);
        float4 P00 = computeSpriteCorner(dir2d, projpos1, scale, float4(-MINSZ, -MAXSZ, -MINSZ, -MAXSZ)); //P - float4(Sz.z, Sz.w, 0, 0);
        output.position = P00;
        output.tc = float2(0,0);
        Stream.Append(output);
        output.position = computeSpriteCorner(dir2d, projpos1, scale, float4(MINSZ, MAXSZ, -MINSZ, -MAXSZ)); //P - float4(Sz.x, Sz.w, 0, 0);
        output.tc = float2(1,0);
        Stream.Append(output);
        Stream.RestartStrip();

    }
}
//////////////////////////////////////////////////////////////////////////////////
// VTX SHADERS VTX SHADERS VTX SHADERS VTX SHADERS VTX SHADERS VTX SHADERS
//////////////////////////////////////////////////////////////////////////////////
Sparkles_VSOut VSSparkles(Sparkles_VSIn input, uint instanceID : SV_InstanceID)
{
    Sparkles_VSOut output;
    float3 P = input.position;
    output.position = float4(P, 1);
    output.normal = input.normal;
    output.instanceID = instanceID;
    return output;
}
//////////////////////////////////////////////////////////////////////////////////
// GEOM. SHADERS GEOM. SHADERS GEOM. SHADERS GEOM. SHADERS GEOM. SHADERS
// A Sprite (quad) will have 4 x (4+3) = 28 floats
// Let's set the Max amount of sprites to 2. You can ask for more but beware that this has a cost (!)
#define MAXVTX      (4)
//////////////////////////////////////////////////////////////////////////////////
[maxvertexcount (MAXVTX)]
void GSSparkles_Barycentric(inout TriangleStream<Sparkles_GSOut> Stream, triangle Sparkles_VSOut input[3], 
                            uint primID : SV_PrimitiveID )
{
    float instanceID = input[0].instanceID; // take one vertex for instanceID...
    float4 projP[3];
    float2 pos2d[3];
    for(int i=0; i<3; i++)
    {
        projP[i] = mul(input[i].position, worldViewProj);
        pos2d[i] = projP[i].xy / projP[i].ww;
    }
    float2 v1 = (pos2d[2]-pos2d[0]);
    float2 v2 = (pos2d[1]-pos2d[0]);
    float projSurf = (v1.x*v2.y - v2.x*v1.y) * LODScale;

    float nsprites = min(projSurf, g_maxSprites);
    
    if(instanceID > nsprites) // This instance has nothering to do here
        return;

    // use primID to add some 'noise' on the way to fetch barycentric coords
    float offset = primID % (256.0 - g_maxSprites);

    if(nsprites >= 1.0)
    {
        nsprites = instanceID + offset;
        Sparkles_VSOut Vtx = barycentricWeights(input[0], input[1], input[2], BCCoords[nsprites]);
        drawSprite(Stream, Vtx, lightPos, 1.0);
    } 
    else if(0 == (primID % round(1.0/projSurf)))// here when we want to bypass some of the triangles
    {
        Sparkles_VSOut Vtx = barycentricWeights(input[0], input[1], input[2], BCCoords[offset]);
        drawSprite(Stream, Vtx, lightPos, 1.0);
    }
}
//////////////////////////////////////////////////////////////////////////////////
// PIXEL SHADERS PIXEL SHADERS PIXEL SHADERS PIXEL SHADERS PIXEL SHADERS
//////////////////////////////////////////////////////////////////////////////////
Sparkles_PSOut PSSparkles(Sparkles_GSOut input)
{
    Sparkles_PSOut output;
    output.color = input.alpha * (texture_star.Sample(sampler_star, input.tc).rrrr);//float4(0,0.5,0.8,1);
    return output;
}

//////////////////////////////////////////////////////////////////////////////////
// TECHNIQUES TECHNIQUES TECHNIQUES TECHNIQUES TECHNIQUES TECHNIQUES TECHNIQUES
//////////////////////////////////////////////////////////////////////////////////

technique10 Sparkles_And_CarPaint
{
    pass ColorNDepth
    {
        SetBlendState(blendOFF, float4(1.0, 1.0, 1.0, 1.0) ,0xffffffff);
        SetRasterizerState( RStateMSAA );
        SetDepthStencilState( depthEnabled, 0 );
        SetVertexShader( CompileShader( vs_4_0, carpaintVS() ) );
        SetGeometryShader(NULL);
        SetPixelShader( CompileShader( ps_4_0, carpaintPS(
            paintamp,
            diffStrength,specStrength,Shininess,
            reflection,
            ambiLightColor,surfColor) ) );
    }
    pass Sparkles
    {
        SetDepthStencilState( depthDisabled, 0 );
        SetRasterizerState( cullDisabled );
        SetBlendState(blendON, float4(1.0, 1.0, 1.0, 1.0) ,0xffffffff);
        SetVertexShader( CompileShader( vs_4_0, VSSparkles() ) );
        SetGeometryShader( CompileShader( gs_4_0, GSSparkles_Barycentric() ) );
        SetPixelShader( CompileShader( ps_4_0, PSSparkles() ) );
    }
}

